
#ifndef __VSIDATA_HPP__
#define __VSIDATA_HPP__
/** 
 * A Class for VSI data Handling
 * @Version $Rev: 58 $
 * @Author M.Sekido
 * @Date  $Date: 2013-07-06 11:58:40 +0900 (Sat, 06 Jul 2013) $
 * $Id: VsiData.hpp 58 2013-07-06 02:58:40Z vlbi $ 
 */

#include <iostream>
#include <vector>
using namespace std;
#include "DAS_info.hpp"

extern "C"{
#include <stdint.h>
}

const int MAX_NBIT = 8;
// const int MAX_NCH = 8;

#ifndef MAX_NCH
#define MAX_NCH (16)
#endif

 
class VsiData{
public:
  /**! Bit encoding order, Same with ADS2000 or its reverse. 
   * ADS2000_2bit-> 00, 01, 10, 11, from bottom to the top level
   * ADS3000_2bit-> 00, 10, 01, 11, from bottom to the top level
   * SignMag_2bit     -> 01, 00, 10, 11, from bottom to the top level
   * RevSignMag_2bit  -> 10, 00, 01, 11, from bottom to the top level
   */
  enum BitEnc_t {ADS2000_2bit, ADS3000_2bit, SignMag_2bit, RevSgnMag_2bit};
  /** Constructor */
  VsiData(DAS_Info* das=0):dasInfo(das){;}
  /** Making look-up table. */
  const int Make_Table(const BitEnc_t t=ADS2000_2bit);
  /** Set DAS information class */
  void Set_DasInfo( DAS_Info* das){ dasInfo = das;  }
  /** Reading 1 sec of data from the file */
  const int32_t Read1Sec(FILE *fp);
  /** Value of the data at (ch,ith) */
  const int8_t Val(const int ch, const int32_t i);
  /** Interpretation (decoding)
   * \return Number of data points.   */
  const long Interpret( const long npt=0);
  /** Interpretation (decoding) 
   * \return Number of data points.   */
  const long Interpret(vector <uint8_t>& datbuf, const long npt);
  /** Returns reference of value array at ch*/
  vector <int8_t>& ValArr(const int ch){ return valarr[ch];}
  /** Returns pointer of value array at ch */
  const int8_t *pValArr(const int ch) {return &valarr[ch][0];} ///< Returns pointer of one
  /** Returns reference of float value array at ch*/
  vector <float>& fValArr(const int ch){ return fvalarr[ch];}
  /** Returns pointer of float value array at ch */
  const float *pfValArr(const int ch) {return &fvalarr[ch][0];} ///< Returns pointer of one
private:
  const int MkTbl1bit(void);
  const int MkTbl2bit(const BitEnc_t t=ADS2000_2bit);
  const int MkTbl4bit(void);
  const int MkTbl8bit(void);
  DAS_Info *dasInfo;
  vector <uint8_t> buf;
  vector <int8_t> valarr[16]; 
  vector <float> fvalarr[16]; 
  int8_t table[256][MAX_NCH];
  float  ftable[256][MAX_NCH];
};

const int8_t level_1bit[2] = {-1,1};
const int8_t level_2bit[4] = {-3, -1, 1, 3 };
const int8_t level_ADS3000_2bit[4] = {-3, 1, -1, 3 };
const int8_t level_SignMag_2bit[4] = {1, -3, 1, 3 };
const int8_t level_RevSgnMag_2bit[4] = {-1, 1, -3, 3};



inline const long VsiData::Interpret(vector <uint8_t>& datbuf, 
				     const long ndat){
  if(dasInfo ==0){  
    cerr << "Error @VsiData::Val():dasInfo is empty.\n";
    return -1;  
  }
	

  int nch= dasInfo->Nch();
  int bit = dasInfo->Qbit();
  int dunit = nch * bit;
  int ch_byte = 8/bit;
  int unit_byte(0);
  int byte_unit (0);
  char msg[256];
  try{
    if( nch > ch_byte){  // 1 unit takes more than 1 Byte
      if( dunit % 8 != 0){
	sprintf(msg, 
		"Nch(%d)*Qbit(%d) must be multiple of 8 bit or less than 8bit", nch, bit);
	throw msg;
      }
      byte_unit = dunit / 8;
    }else{  // 1 byte include no less than 1 unit
      if( 8 % dunit != 0 ){
	sprintf(msg, "Nch(%d)*Qbit(%d) must be multiple of 8 bit or less than 8bit", nch, bit);
	throw msg;
      }else{
	unit_byte = 8 / dunit;
      }
    }
  } 
  catch(const char* msg1){
    cerr << "Error in VsiData::Val():" << msg1 << endl;
    return -1;
  }
  size_t npt = ndat;
  if(ndat == 0) npt = dasInfo->Srate();
  for( int i=0; i<nch; i++) 
    if( valarr[i].size() < npt){
      valarr[i].resize(npt);
      fvalarr[i].resize(npt);
    }
  vector <uint8_t>::iterator sptr = datbuf.begin();  
  vector <float>::iterator dptr[nch];
  vector <int8_t>::iterator dptr2[nch];

  for(int i=0; i<nch; i++){
    dptr[i] = fvalarr[i].begin();  
    dptr2[i] = valarr[i].begin();  
  }

  int32_t cnt=0;
  
  if( byte_unit == 0){ // 1 byte contains multiple units
    while(sptr < datbuf.end() && dptr[0] < fvalarr[0].end()){
      for( int i=0; i< unit_byte; i++){
	for(int j=0; j<nch; j++){
	  *dptr[j] = ftable[*sptr][j+i*nch]; 
	  *dptr2[j] = table[*sptr][j+i*nch]; 
	  dptr[j]++; dptr2[j]++;
	}
	cnt++;
      }
      sptr++;
    } 
  }else{ // 1 unit takes multiple bytes
    while(sptr < datbuf.end() && dptr[0] < fvalarr[0].end()){
      int  ich=0;
      for(int i=0; i<byte_unit; i++){ // bytes/unit
	for(int jch=0; jch< ch_byte; jch++,ich++){ // ch/byte
	  *dptr[ich] = ftable[*sptr][jch];  
	  *dptr2[ich] = table[*sptr][jch];  
	  dptr[ich]++; 	  dptr2[ich]++; 
	}
	sptr++;
      }
      cnt++;
    }
  } 
  
  return cnt;
}


inline const long VsiData::Interpret( const long ndat){
  if(dasInfo ==0){  
    cerr << "Error @VsiData::Val():dasInfo is empty.\n";
    return -1;  
  }
  
  int nch= dasInfo->Nch();
  int bit = dasInfo->Qbit();
  int dunit = nch * bit;
  int ch_byte = 8/bit;
  int unit_byte(0);
  int byte_unit (0);
  char msg[256];
  try{
    if( nch > ch_byte){  // 1 unit takes more than 1 ch
      if( dunit % 8 != 0){
	sprintf(msg, 
		"Nch(%d)*Qbit(%d) must be multiple of 8 bit or less than 8bit", nch, bit);
	throw msg;
      }
      byte_unit = dunit / 8;
    }else{  // 1 byte include no less than 1 channel
      if( 8 % dunit != 0 ){
	sprintf(msg, "Nch(%d)*Qbit(%d) must be multiple of 8 bit or less than 8bit", nch, bit);
	throw msg;
      }else{
	unit_byte = 8 / dunit;
      }
    }
  } 
  catch(const char* msg1){
    cerr << "Error in VsiData::Val():" << msg1 << endl;
    return -1;
  }
  
  size_t npt = ndat;
  if(ndat == 0)  npt = dasInfo->Srate();
  for( int i=0; i<nch; i++) 
    if( valarr[i].size() < npt) valarr[i].resize(npt);
  vector <uint8_t>::iterator sptr = buf.begin();  
  vector <float>::iterator dptr[nch];
  vector <int8_t>::iterator dptr2[nch];
  for(int i=0; i<nch; i++){
    dptr[i] = fvalarr[i].begin();  
    dptr2[i] = valarr[i].begin();  
  }

  int32_t cnt=0;
  
  if( byte_unit == 0){ // 1 byte contains multiple units
    while(sptr < buf.end() && dptr2[0] < valarr[0].end()){
      for( int i=0; i< unit_byte; i++){
	for(int j=0; j<nch; j++){
	  *dptr2[j] = table[*sptr][j+i*nch]; 
	  *dptr[j] = ftable[*sptr][j+i*nch]; 
	  dptr[j]++;	  dptr2[j]++;
	}
	cnt++;
      }
      sptr++;
    } 
  }else{ // 1 unit takes multiple bytes
    while(sptr < buf.end() && dptr2[0] < valarr[0].end()){
      int  ich=0;
      for(int i=0; i<byte_unit; i++){ // bytes/unit
	for(int jch=0; jch< ch_byte; jch++,ich++){ // ch/byte
	  *dptr2[ich] = table[*sptr][jch];  
	  *dptr[ich] = ftable[*sptr][jch];  
	  dptr[ich]++;	  dptr2[ich]++;
	}
	sptr++;
      }
      cnt++;
    }
  } 
  
  return cnt;
}



inline const int8_t VsiData::Val(const int ch, const int32_t ith){
  if(dasInfo ==0){  
    cerr << "Error @VsiData::Val():dasInfo is empty.\n";
    return -100;  
  }
  int nch= dasInfo->Nch();
  int bit = dasInfo->Qbit();
  int dunit = nch * bit;
  int ch_byte = 8/bit;
  int unit_byte(0);
  int byte_unit (0);
  char msg[256];

  try{
    if( nch > ch_byte){  // 1 unit takes more than 1 ch
      if( dunit % 8 != 0){
	sprintf(msg, "Nch(%d)*Qbit(%d) must be multiple of 8 bit or less than 8bit", nch, bit);
	throw msg;
      }
      byte_unit = dunit / 8;
    }else{  // 1 byte include no less than 1 channel
      if(8 % dunit != 0){
	sprintf(msg, "Nch(%d)*Qbit(%d) must be multiple of 8 bit or less than 8bit", nch, bit);
	throw msg;
      }else{
	unit_byte = 8/dunit;
      }
    }
  } 
  catch(const char* msg){
    cerr << "Error in VsiData::Val():" << msg << endl;
    return -100;
  }

  if( byte_unit == 0 ){
    int32_t ibyte = ith / unit_byte;
    int frac_unit = ith %  unit_byte;
    return table[buf[ibyte]][ch+frac_unit*nch];
  }else{
    int32_t ibyte = ith*byte_unit;
    int abyte = ch /ch_byte;
    ibyte += abyte;
    int ich = ch % ch_byte;
    return table[buf[ibyte]][ich];
  }
  
}

#endif
